import dis
import importlib
import builtins
import inspect

from pycarol.pipeline.utils import int_to_bytes, flat_list

VERBOSE = False  # dev parameter


def get_consts_hash(f) -> bytes:
    if hasattr(f, '__code__'):  # is function object
        consts_list = list(f.__code__.co_consts)
    elif hasattr(f, 'co_code'):  # is code object
        consts_list = list(f.co_consts)
    else:
        raise TypeError

    for i, v in enumerate(consts_list):
        if isinstance(v, str) and "<locals>" in v:
            consts_list[i] = v.split('.')[-1]
    consts_tuple = tuple(consts_list)
    return int_to_bytes(hash(consts_tuple))


number_of_parameters_in_build_ops = dict(
    BUILD_TUPLE=lambda x: x,
    BUILD_LIST=lambda x: x,
    BUILD_SET=lambda x: x,
    BUILD_MAP=lambda x: 2 * x,
    BUILD_CONST_KEY_MAP=lambda x: x + 1,
    BUILD_STRING=lambda x: x,
    BUILD_TUPLE_UNPACK=lambda x: x,
    BUILD_TUPLE_UNPACK_WITH_CALL=lambda x: x,
    BUILD_LIST_UNPACK=lambda x: x,
    BUILD_SET_UNPACK=lambda x: x,
    BUILD_MAP_UNPACK=lambda x: x,
    BUILD_MAP_UNPACK_WITH_CALL=lambda x: x,
)


def find_called_function(ix, inst, instructions):
    """
    When a CALL_FUNCTION instruction is found, the function name is not given
    as a direct argument to this instruction. Instead, the function name can
    be found some instructions above on the bytecode. Between the CALL_FUNCTION
    instruction and the function pointer we found all the function parameters.
    This method implements the logic needed to fetch the function pointer for
    the three kind of CALL_FUNCTION operations.

    Args:
        ix: index of call function instruction
        inst: call function instruction
        instructions: list of instructions composing the whole bytecode

    Returns:
        function_name: the name of the called function
    """
    if "CALL_FUNCTION" == inst.opname:  # it is simple call function instruction
        # for this instruction, we can find the called function some instructions
        # above. we just need to skip backwards the number of arguments
        offset = inst.arg + 1
        called_function_inst = instructions[ix - offset]
    elif "CALL_FUNCTION_KW" == inst.opname:  # call function op with keyword arguments
        # wrt CALL_FUNCTION there is one additional argument to skip
        offset = inst.arg + 2
        called_function_inst = instructions[ix - offset]
    elif "CALL_FUNCTION_EX":
        offset = inst.arg + 2
        # Next, we look for BUILD instructions between CALL_FUNCTION_EX instruction
        # and estimated target function position defined by offset. Then, we update
        # offset following each instruction behavior to skip all the parameters
        # and finally find the target function name.
        rev_ind = 0
        while rev_ind < offset:
            rev_ind += 1
            ind_i = ix - rev_ind
            if ind_i < 0:
                raise KeyError("step out of the bytecode while searching CALL_FUNCION_EX target")
            inst_i = instructions[ind_i]
            opname = inst_i.opname
            arg = inst_i.arg
            if opname in number_of_parameters_in_build_ops:  # increase offset
                offset += number_of_parameters_in_build_ops[opname](arg)
        called_function_inst = instructions[ix - offset]
    else:
        raise NotImplementedError("instruction {} is not supported".format(inst.opname))
    if VERBOSE:
        print(called_function_inst)
        print("offset: ", offset)

    function_name = called_function_inst.argval
    return function_name


def find_defined_function(ix, inst, instructions):
    """
    When a MAKE_FUNCTION instruction is found, the name and code pointer can
    be found on the two instructions above.
    Args:
        ix: index of MAKE_FUNCTION instruction
        inst: MAKE_FUNCION instruction
        instructions: list of instructions composing the whole bytecode

    Returns: a dict, whose only key is the function name, and value is function code

    """
    assert inst.opname == "MAKE_FUNCTION"
    assert ix >= 2

    name = instructions[ix - 1].argval.split('.')[-1]  # discard context namespace
    code = instructions[ix - 2].argval

    return {name: code}


def fetch_function_object(function_name: str, enclosing_function, local_functions: dict):
    """
    function_name can be found in Local, Enclosed, Global and Builtin Namespace
    Local and Enclosed namespaces cannot be exactly inferred in static analysis.
    For estimating local and enclosed namespaces, we do not consider conditional
    branches(jumps) in the bytecode. All functions definitions encountered are
    evaluated in the order they appear in the bytecode.

    Args:
        function_name: name of called function
        enclosing_function (function object or bytecode): context function in which
        call function happens
        local_functions: dict containing functions defined in enclosing_function

    Returns:
        func: function object
    """

    if function_name in local_functions:  # Local Namespace
        return local_functions[function_name]

    if hasattr(enclosing_function, '__module__'):
        module_namespace = importlib.import_module(enclosing_function.__module__)
        if hasattr(module_namespace, function_name):  # Module (Global) Namespace
            return getattr(module_namespace, function_name)

    if hasattr(builtins, function_name):  # Builtin Namespace
        return getattr(builtins, function_name)
    # else
    raise NotImplementedError("{} was neither found in module {} nor it is builtin.\
    \nLocal and Enclosed namespaces are not supported".format(function_name, module_namespace))


def get_bytecode_tree(top_function: 'function', ignore_not_found_function=False) -> list:
    """

    Args:
        top_function:
        ignore_not_found_function:

    Returns:

    """

    def _traverse_code(parent_function: 'function') -> list:

        if hasattr(parent_function, '__name__'):
            if hasattr(builtins, parent_function.__name__):
                # dis cannot get instructions for builtins
                # return function name instead of bytecode
                return [parent_function.__name__]

        nonlocal code_set
        called_functions = set()
        locally_defined_functions: dict = {}
        instructions = list(dis.get_instructions(parent_function))

        for ix, inst in enumerate(instructions):
            context = (ix, inst, instructions)

            if inst.opname == "MAKE_FUNCTION":  # locally defined function
                locally_defined_functions.update(find_defined_function(*context))

            if "CALL_FUNCTION" in inst.opname:  # call_function op found
                son_function_name = find_called_function(*context)
                try:
                    son_function = fetch_function_object(
                        son_function_name, parent_function, locally_defined_functions
                    )
                except NotImplementedError as e:
                    if ignore_not_found_function:
                        continue
                    else:
                        raise e
                if son_function not in code_set:
                    code_set.add(son_function)
                    called_functions.add(son_function)

        code_list = [
            _traverse_code(f) for f in called_functions
            # if f not in code_set  # analyze each function only once
        ]

        if hasattr(parent_function, '__code__'):  # parent_function is function object
            consts = parent_function.__code__.co_consts


            function_code: list = b''.join([
                parent_function.__code__.co_code,
                get_consts_hash(parent_function),
                int_to_bytes(hash(
                    dict(inspect.getmembers(parent_function))['__defaults__']
                )),
            ])
        elif hasattr(parent_function, 'co_code'):  # parent_function is code object
            function_code: list = b''.join([
                parent_function.co_code,
                get_consts_hash(parent_function),
                # missing default parameter information
                # asbytes(hash(
                #     dict(inspect.getmembers(parent_function))['__defaults__']
                # )),
            ])
        else:
            raise TypeError("parent_function should be either function or code.")

        code_list.append(function_code)
        return code_list

    if not hasattr(top_function, '__code__'):
        raise TypeError('argument should be a function')

    code_set = set()  # memoization set
    bytecode_tree = _traverse_code(top_function)
    assert isinstance(bytecode_tree, list)
    return bytecode_tree


def get_function_hash(f: 'function', ignore_not_found_function=False) -> int:
    """
    Module main function. It returns a proper hash for the given function.
    Args:
        f: function to be hashed
        ignore_not_found_function: setting to True will ignore some errors
        and compute a hash anyway. Some part of the code may be ignored.

    Returns:
        h: a hash number for the given function

    """
    bytecode_nested_list = get_bytecode_tree(f, ignore_not_found_function)
    bytecode_flat_list = flat_list(bytecode_nested_list)
    h = hash(tuple(bytecode_flat_list))
    return h




# TODO: improve documentation
# TODO: real scenario test case
# TODO: support inner imports. possible?



from ..utils import get_reverse_dag, breadth_first_search


def nodes_layout(dag:dict, align_on_leafs = True) -> dict:
    """
    Builds basic graph plot layout. In this version, nodes are placed on x
    axis accordingly to their depth in dag. Top levels are on left whereas
    deeper nodes are placed on the right. If align on_leafs is set to true,
    DAG is first reversed, so that the output of the pipeline is on the right
    and inputs on the left.
    Args:
        dag: dict encoding a DAG structure
        align_on_leafs: layout boolean parameter

    Returns:
        layout: a dict whose keys are DAG nodes an values are (x,y) of each
        node.

    """

    layout = {}
    if align_on_leafs:
        dag = get_reverse_dag(dag)
    for i, nodes in enumerate(breadth_first_search(dag)):
        for j, node in enumerate(nodes):
            layout[node] = (i,j)
    return layout

def edges_layout(dag:dict, layout:dict) -> list:
    """
    Given a dag network and the positions of each one of its nodes,
    this function creates the positions of the edges.
    Args:
        dag: dict encoding a DAG
        layout: dict containing nodes x,y positions

    Returns:
        edges: list of positions, each one in format ( (x0,y0), (x1,y1) )
    """

    edges = []
    for source_node in layout:
        for target_node in dag[source_node]:
            edges.append((layout[source_node],layout[target_node]),)
    return edges

def _get_task_id(t):
    return t.task_id

def _get_task_family(t):
    if '.' in t.task_id:
        return t.task_id.split('.')[-2]
    else:
        return "empty_namespace"

def _get_task_name(t):
    return t.task_id.split('.')[-1]

def _get_complete(t):
    return t.complete()

def _get_tasklog(t):
    log = t.loadlog()
    if isinstance(log, str):
        return log
    else:
        return "Log type is wrong."

def _get_hash_version(t):
    return ""


def make_nodes_data_source(nodes_layout) -> dict:
    """
    Creates a bokeh compatible data source encoding nodes plotting
    properties. Returns this data source in bokeh compatible dict format
    Args:
        nodes_layout: dict containing nodes as keys and its positions as
        values. In this function nodes are luigi tasks.

    Returns:
        data_source: bokeh compatible dict containing the columns: x, y,
        task_id, task_family, task_name, complete, tasklog, hash_version

    """

    data = dict(
        # task=[],
        x=[],
        y=[],
        task_id=[],
        task_family=[],
        task_name=[],
        complete=[],
        tasklog=[],
        hash_version=[],
    )
    for k,(x,y) in nodes_layout.items():
        # data['task'].append(k)
        data['x'].append(x)
        data['y'].append(y)
        data['task_id'].append(_get_task_id(k))
        data['task_family'].append(_get_task_family(k))
        data['task_name'].append(_get_task_name(k))
        data['complete'].append(_get_complete(k))
        data['tasklog'].append(_get_tasklog(k))
        data['hash_version'].append(_get_hash_version(k))

    return data

def make_edges_data_source(edges_layout) -> dict:
    """
    Creates a bokeh segment glyph compatible data source encoding edges plotting
    properties. Returns this data source in bokeh compatible dict format
    Args:
        edges_layout: list containing edges coordinates.

    Returns:
        data_source: bokeh segment glyph compatible dict containing the
        columns: x0, y0, x1, y1

    """

    data = dict(
        x0=[],
        y0=[],
        x1=[],
        y1=[]
    )
    for ((x0,y0,),(x1,y1)) in edges_layout:
        data['x0'].append(x0)
        data['y0'].append(y0)
        data['x1'].append(x1)
        data['y1'].append(y1)

    return data




from pycarol.pipeline import Task
from pycarol.pipeline.utils import build_dag

def luigi_get_sons(task) -> list:
    """
    Returns a list of required tasks. This is used in build_dag
    Args:
        task: luigi Task

    Returns:
        l: list of luigi Task

    """
    return task.requires()


def get_dag_from_task(task:list) -> dict:
    """
    Wrapper around generic build_dag.
    Args:
        task: list of proper luigi tasks

    Returns:
        dag: dict encoding a DAG

    """
    dag = build_dag(task,luigi_get_sons)
    return dag


class Pipe(object):
    def __init__(self, tasks: list):
        assert isinstance(tasks,list)

        for t in tasks:
            assert isinstance(t,Task)
        self.top_nodes = tasks
        self.dag = get_dag_from_task(tasks)
        self.all_tasks = [k for k in self.dag]
    def remove_upstream(self, task):
        return
    def remove_all(self):
        return
    def describe_targets(self):
        return
    def get_task_by_name(self,name):
        return
    def get_dag(self):
        return self.dag


import bokeh
assert bokeh.__version__ == '1.2.0'
from bokeh.models import (
    ColumnDataSource,
    CDSView,
    BooleanFilter,
    HoverTool,
)

from bokeh.events import ButtonClick
from bokeh.models import Button
from bokeh.layouts import widgetbox
from bokeh.models.widgets import PreText

from bokeh.layouts import column, row, layout
from bokeh.plotting import figure
from bokeh.transform import transform
from bokeh.palettes import Category10
from bokeh.models.mappers import CategoricalColorMapper

def _make_colormapper(data_source: dict, col_name: str):
    import numpy as np
    factors = list(set(data_source[col_name]))
    nb_factors = np.clip(len(factors), 3, 10)
    return CategoricalColorMapper(factors=factors,
                                  palette=Category10[nb_factors])


def _make_pipeline_plot(
        nodes_data_source,
        edges_data_source,
):

    family_color = _make_colormapper(nodes_data_source.data,'task_family')

    pipeline_plot = figure(
        title="Pipeline Debugger",
        x_range=(-1, max(nodes_data_source.data['x']) + 1),
        y_range=(-1, max(nodes_data_source.data['y']) + 1),
        tools=[
            HoverTool(names=['alltasks']),
            'wheel_zoom',
            'pan',
            'tap',
            'box_select',
            'reset'
        ],
        toolbar_location='right',
        tooltips=[
            ('index', "$index"),
            ('task_id', '@task_id'),
        ],
        plot_width=900,
        plot_height=500,
    )

    edges_glyph = pipeline_plot.segment(
        'x0',
        'y0',
        'x1',
        'y1',
        source=edges_data_source,
        color='gray',
        line_alpha=0.3,
    )

    pipeline_plot.circle(
        'x',
        'y',
        source=nodes_data_source,
        size=20,
        name='alltasks',
        color=transform('task_family', family_color),
        legend='task_family',
    )

    # pipeline_plot.circle(
    #     'x',
    #     'y',
    #     source=s2,
    #     size=10,
    #     color='white',
    # )

    pipeline_plot.text(
        x='x',
        y='y',
        text='task_name',
        source=nodes_data_source,
        text_font_size='8pt',
    )

    edges_glyph.nonselection_glyph = None
    return pipeline_plot

class PlotDynamics():
    def __init__(
            self,
            nodes_data_source,
            edges_data_source,
    ):
        nodes_data_source.selected.on_change(
            'indices', self.select_callback
        )

        self.remove_button = Button(label='Remove Selected')
        self.removeupstream_button = Button(label='Remove Upstream')
        self.update_button = Button(label='Update')

        self.remove_button.on_event(
            ButtonClick,
            self.remove_callback,
        )
        self.removeupstream_button.on_event(
            ButtonClick,
            self.removeupstream_callback,
        )
        self.update_button.on_event(
            ButtonClick,
            self.update_callback,
        )

        self.nodes_data_source = nodes_data_source
        self.edges_data_source = edges_data_source
        self.selected_nodes = []


    def select_callback(self,attr, old, new):
        self.selected_nodes = new
        # pre.text = self.tasklog[new[0]]
        print("selected {}".format(new))

    def remove_callback(self,event):
        return

        # target_list = self.nodes_data_source.selected.indices
        # for t in target_list:
        #     print('removing', t)
        #     task = self.
        #     TaskViewer.nodes[t].remove()

    def removeupstream_callback(self,event):
        return

    def update_callback(event):
        # complete = self.update_complete()
        # s2.data = {k: [vi for i, vi in enumerate(v) if not complete[i]] for
        #            k, v in source_dict.items()}
        return
    def buttons(self):
        return row([
            self.remove_button,
            self.update_button,
            self.removeupstream_button,
        ])

def plot_pipeline(nodes_data,edges_data,pipe):
    """
    Receives data sources in dict format and returns dynamic plot
    Args:
        nodes_data:
        edges_data:

    Returns:
        final_layout:

    """

    nodes_data_source = ColumnDataSource(data=nodes_data)
    edges_data_source = ColumnDataSource(data=edges_data)

    # done_tasks_data_source = CDSView(
    #     source=nodes_data_source,
    #     filters=[BooleanFilter(complete)]
    # )
    # notdone_tasks_data_source = CDSView(
    #     source=nodes_data_source,
    #     filters=[BooleanFilter([not c for c in complete])]
    # )

    pipeline_plot = _make_pipeline_plot(nodes_data_source, edges_data_source)

    dynamics = PlotDynamics(nodes_data_source,edges_data_source)

    ### Layout
    final_layout = layout(
        column(
            [
                pipeline_plot,
                row(dynamics.buttons()),
            ],
            sizing_mode='scale_width'
        ),
    )
    return final_layout


def get_plot_from_pipeline(pipe):
    """
    Main module method. From a luigi task with defined parameters, generates a
    bokeh plot of the
    pipeline. It does not render the plot.
    Args:
        pipe: list of luigi task initialised with proper parameters

    Returns:
        bokeh_layout:

    """

    from .viewer import (nodes_layout, edges_layout,
                         make_nodes_data_source, make_edges_data_source,
                         )

    dag = pipe.get_dag()

    nodes_layout = nodes_layout(dag)
    edges_layout = edges_layout(dag, nodes_layout)

    nodes_data_source = make_nodes_data_source(nodes_layout)
    edges_data_source = make_edges_data_source(edges_layout)

    bokeh_layout = plot_pipeline(nodes_data_source, edges_data_source,pipe)

    return bokeh_layout






import luigi
from luigi import parameter, six
from luigi.task import flatten
from pycarol.pipeline.targets import PickleTarget
import logging
import warnings

logger = logging.getLogger('luigi-interface')
logger.setLevel(logging.INFO)


class Task(luigi.Task):
    TARGET_DIR = './luigi_targets/'  # this class attribute can be redefined somewhere else

    TARGET = PickleTarget  # DEPRECATED!
    target_type = PickleTarget
    is_cloud_target = None

    persist_stdout = False
    requires_list = []
    requires_dict = {}
    resources = {'cpu': 1}  # default resource to be overridden or complemented

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def visualize(self):
        # override this method to provide a visualization for taskviewer
        return None

    def buildme(self, local_scheduler=True, **kwargs):
        luigi.build([self, ], local_scheduler=local_scheduler, **kwargs)

    def debug_task(self):
        persist_stdout =self.persist_stdout
        self.persist_stdout = False
        self.run()
        self.persist_stdout = persist_stdout

    def _file_id(self):
        # returns the output default file identifier
        return luigi.task.task_id_str(self.get_task_family(), self.to_str_params(only_significant=True))

    def _txt_path(self):
        return "{}.txt".format(self._file_id())

    def requires(self):
        if len(self.requires_list) > 0:
            result_list = []
            for t in self.requires_list:
                fixed_params = {}
                if type(t) is tuple:
                    fixed_params = t[1]
                    t = t[0]
                task_instance = self.clone(t, **fixed_params)
                result_list.append(task_instance)
            return result_list
        elif len(self.requires_dict) > 0:
            result_dict = {}
            for k, t in self.requires_dict.items():
                fixed_params = {}
                if type(t) is tuple:
                    fixed_params = t[1]
                    t = t[0]
                task_instance = self.clone(t, **fixed_params)
                result_dict.update({k: task_instance})
            return result_dict
        else:
            return []

    def output(self):
        if self.TARGET != PickleTarget:  # Check for deprecated use
            warnings.warn('TARGET is being replaced with target_type.', DeprecationWarning)
            return self.TARGET(self)

        return self.target_type(self)

    def load(self, **kwargs):
        return self.output().load(**kwargs)

    def remove(self):
        self.output().remove()
        self.output().removelog()

    def loadlog(self):
        return self.output().loadlog()

    def run(self):

        if isinstance(self.input(), list):
            function_inputs = [input_i.load(**self.load_input_params(input_i)) if self.load_input_params(input_i)
                               else input_i.load() for input_i in self.input()]
        elif isinstance(self.input(), dict):
            function_inputs = {i: (input_i.load(**self.load_input_params(input_i))
                                   if self.load_input_params(input_i) else input_i.load())
                               for i, input_i in self.input().items()}


        if hasattr(self.output(), '_is_cloud_target') and self.output()._is_cloud_target and self.persist_stdout:
            try:
                from contextlib import redirect_stdout
                import io
                f = io.StringIO()
                with redirect_stdout(f):
                    function_output = self._easy_run(function_inputs)
            finally:
                # self.output().persistlog(self._txt_path())
                self.output().persistlog(f.getvalue())
        else:
            function_output = self._easy_run(function_inputs)

        print("dumping task", self)

        self.output().dump(function_output)

    def _easy_run(self, inputs):
        # Override this method to implement standard pre/post-processing
        return self.easy_run(inputs)

    def easy_run(self, inputs):
        return None

    #this method was changed from the original version to allow execution of a task
    #with extra parameters. the original one, raises an exception. now, we print that exception
    #in this version we do not raise neither print it.
    @classmethod
    def get_param_values(cls, params, args, kwargs):
        """
        Get the values of the parameters from the args and kwargs.
        :param params: list of (param_name, Parameter).
        :param args: positional arguments
        :param kwargs: keyword arguments.
        :returns: list of `(name, value)` tuples, one for each parameter.
        """
        result = {}

        params_dict = dict(params)

        task_family = cls.get_task_family()

        # In case any exceptions are thrown, create a helpful description of how the Task was invoked
        # TODO: should we detect non-reprable arguments? These will lead to mysterious errors
        exc_desc = '%s[args=%s, kwargs=%s]' % (task_family, args, kwargs)

        # Fill in the positional arguments
        positional_params = [(n, p) for n, p in params if p.positional]
        for i, arg in enumerate(args):
            if i >= len(positional_params):
                raise parameter.UnknownParameterException(
                    '%s: takes at most %d parameters (%d given)' % (exc_desc, len(positional_params), len(args)))
            param_name, param_obj = positional_params[i]
            result[param_name] = param_obj.normalize(arg)

        # Then the keyword arguments
        for param_name, arg in six.iteritems(kwargs):
            if param_name in result:
                raise parameter.DuplicateParameterException(
                    '%s: parameter %s was already set as a positional parameter' % (exc_desc, param_name))
            if param_name not in params_dict:
                # raise parameter.UnknownParameterException('%s: unknown parameter %s' % (exc_desc, param_name))
                continue

            result[param_name] = params_dict[param_name].normalize(arg)

        # Then use the defaults for anything not filled in
        for param_name, param_obj in params:
            if param_name not in result:
                if not param_obj.has_task_value(task_family, param_name):
                    raise parameter.MissingParameterException(
                        "%s: requires the '%s' parameter to be set" % (exc_desc, param_name))
                result[param_name] = param_obj.task_value(task_family, param_name)

        def list_to_tuple(x):
            """ Make tuples out of lists and sets to allow hashing """
            if isinstance(x, list) or isinstance(x, set):
                return tuple(x)
            else:
                return x

        # Sort it by the correct order and make a list
        return [(param_name, list_to_tuple(result[param_name])) for param_name, param_obj in params]

    def get_execution_params(self):
        params = {}
        for param_name in dir(self.__class__):
            param_obj = getattr(self, param_name)
            params.update({param_name: param_obj})

        return params

    def load_input_params(self, input_target):
        """
        Overwrite this if need to pass parameters when loading a requirement.

        :param input_target:
            Target that will be loaded.
        :return: `Dict`
            Return a dict with key/value parameters to be passed to Target.load()
        """
        return {}


class WrapperTask(Task):
    """
    Use for tasks that only wrap other tasks and that by definition are done if all their requirements exist.
    """
    def run(self):
        pass

    def complete(self):
        return all(r.complete() for r in flatten(self.requires()))

    def output(self):
        return self.input()


def set_attributes(task_to_inherit, task_that_inherits):
    fixed_params = {}

    if type(task_to_inherit) is tuple:
        fixed_params = task_to_inherit[1]
        task_to_inherit = task_to_inherit[0]

    for param_name, param_obj in task_to_inherit.get_params():
        if param_name in fixed_params:  # do not inherit fixed params
            continue

        # Check if the parameter exists in the inheriting task
        if not hasattr(task_that_inherits, param_name):
            # If not, add it to the inheriting task
            setattr(task_that_inherits, param_name, param_obj)
    return task_that_inherits


class inherit_list(object):
    # http://blog.thedigitalcatonline.com/blog/2015/04/23/python-decorators-metaprogramming-with-style/

    def __init__(self, *task_to_inherit_list):
        self.requires_list = list(task_to_inherit_list)

    def __call__(self, task_that_inherits):
        task_that_inherits.requires_list = self.requires_list
        for task_to_inherit in task_that_inherits.requires_list:
            # Get all parameter objects from the underlying task
            task_that_inherits = set_attributes(task_to_inherit, task_that_inherits)

        return task_that_inherits


class inherit_dict(object):
    def __init__(self, **task_to_inherit_dict):
        self.requires_dict = task_to_inherit_dict

    def __call__(self, task_that_inherits):
        task_that_inherits.requires_dict = self.requires_dict

        for key, task_to_inherit in task_that_inherits.requires_dict.items():
            # Get all parameter objects from the underlying task
            task_that_inherits = set_attributes(task_to_inherit, task_that_inherits)
        return task_that_inherits



import warnings
import pandas as pd
import json
from ..utils.importers import _import_dask, _import_pandas
from ..storage import Storage
from ..query import Query
from ..filter import TYPE_FILTER, Filter
import itertools


class DataModelView:

    def __init__(self, carol):
        self.carol = carol


    def get_by_name(self, view_name):
        return self._get(view_name, by='name')


    def _get(self, id,  by='name'):

        if by == 'name':
            url = f"v1/relationshipView/name/{id}"
        elif by == 'id':
            url = f"/api/v1/relationshipView/{id}"
        else:
            raise print('Type incorrect, it should be "id" or "name"')

        return self.carol.call_api(url, method='GET')

    def _build_query_params(self):
        if self.sort_by is None:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order}
        else:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order,
                                 "sortBy": self.sort_by}

    def reprocess(self, view_name=None, view_id=None):

        if view_name is not None:
            view_id = self.get_by_name(view_name=view_name)['mdmId']
        else:
            assert view_id is not None, "'view_id' or 'view_name' must be set"

        url_filter = "v1/goldenRecordView/reprocess"

        query_params = {"relationshipViewId":view_id}
        return self.carol.call_api(url_filter, params=query_params,
                                   method='POST')

    def get_all(self, offset=0, page_size=-1, sort_order='ASC',
                sort_by=None, print_status=False,
                save_file=None):


        self.offset = offset
        self.page_size = page_size
        self.sort_order = sort_order
        self.sort_by = sort_by
        self._build_query_params()

        self.template_dict = {}
        self.template_data = []
        count = self.offset

        set_param = True
        self.total_hits = float("inf")
        if save_file:
            assert isinstance(save_file, str)
            file = open(save_file, 'w', encoding='utf8')
        while count < self.total_hits:
            url_filter = "v1/relationshipView"
            query = self.carol.call_api(url_filter, params=self.query_params, method='GET')

            if query['count'] == 0:
                print('There are no more results.')
                print('Expecting {}, reponse = {}'.format(self.total_hits, count))
                break
            count += query['count']
            if set_param:
                self.total_hits = query["totalHits"]
                set_param = False

            query = query['hits']
            self.template_data.extend(query)

            self.template_dict.update({i['mdmName']: {'mdmId': i['mdmId'],
                                                      'mdmRunningState': i['mdmRunningState'],
                                                      'mdmEntityType': i['mdmEntityType']}
                                       for i in query})

            self.query_params['offset'] = count
            if print_status:
                print('{}/{}'.format(count, self.total_hits), end='\r')
            if save_file:
                file.write(json.dumps(query, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if save_file:
            file.close()
        return self


    def export(self, view_name=None, view_id=None, sync_view=True, full_export=False,
               delete_previous=False, export_format='PARQUET'):
        """

        Export datamodel to s3

        This method will trigger or pause the export of the data in the datamodel to
        s3

        :param view_name: `str`, default `None`
            View Name
        :param view_id: `str`, default `None`
            View id
        :param sync_view: `bool`, default `True`
            Sync the view
        :param full_export: `bool`, default `True`
            Do a resync of the view
        :param delete_previous: `bool`, default `False`
            Delete previous exported files.
        :param export_format: `str`, default `PARQUET`
            Format of the data to be exported. Possible values are:
                "PARQUET", "CSV", "JSON", "EXCEL"
        :return: None
        """

        if sync_view:
            status = 'RUNNING'
        else:
            status = 'PAUSED'

        if view_name:
            view_id = self.get_by_name(view_name)['mdmId']
        else:
            assert view_id

        query_params = {"status": status, "fullExport": full_export,
                        'relationshipViewId':view_id, 'format':export_format,
                        "deletePrevious": delete_previous}

        url = f'v1/goldenRecordView/exporter'
        return self.carol.call_api(url, method='POST', params=query_params)


    def fetch_parquet(self, view_name, merge_records=True, backend='pandas', return_dask_graph=False,
                      columns=None, return_metadata=False, callback=None, max_hits=None):

        """

        :param view_name: `str`
            View name to be imported
        :param merge_records: `bool`, default `True`
            This will keep only the most recent record exported. Sometimes there are updates and/or deletions and
            one should keep only the last records.
        :param backend: ['dask','pandas'], default `dask`
            if to use either dask or pandas to fetch the data
        :param return_dask_graph: `bool`, default `false`
            If to return the dask graph or the dataframe.
        :param columns: `list`, default `None`
            List of columns to fetch.
        :param return_metadata: `bool`, default `False`
            To return or not the fields ['mdmId', 'mdmCounterForEntity']
        :param callback: `callable`, default `None`
            Function to be called each downloaded file.
        :param max_hits: `int`, default `None`
            Number of records to get.
        :return:
        """

        if isinstance(columns, str):
            columns = [columns]

        assert backend == 'dask' or backend == 'pandas'

        if return_dask_graph:
            assert backend == 'dask'


        dms = self._get_view_export_stats()
        if not dms.get(view_name):
            raise Exception(
                f'"{view_name}" is not set to export data, \n'
                f'use `dm = DataModelView(login).export(view_name="{view_name}", sync_dm=True) to activate')

        if columns:
            columns.extend(['mdmId', 'mdmCounterForEntity', 'mdmLastUpdated'])

        storage = Storage(self.carol)
        if backend == 'dask':
            d = _import_dask(storage=storage, view_name=view_name, import_type='view',
                             merge_records=merge_records, return_dask_graph=return_dask_graph,
                             columns=columns)

        elif backend == 'pandas':

            d = _import_pandas(storage=storage, view_name=view_name, golden=True, columns=columns, callback=callback,
                               max_hits=max_hits,import_type='view')
            if d is None:
                warnings.warn("No data to fetch!", UserWarning)
                _field_types = self._get_name_type_DMs(self.get_by_name(dm_name)['mdmFields'])
                cols_keys = list(_field_types)
                if return_metadata:
                    cols_keys.extend(['mdmId', 'mdmCounterForEntity', 'mdmLastUpdated'])

                elif columns:
                    columns = [i for i in columns if i not in ['mdmId', 'mdmCounterForEntity', 'mdmLastUpdated']]

                d = pd.DataFrame(columns=cols_keys)
                for key, value in _field_types.items():
                    d.loc[:, key] = d.loc[:, key].astype(_DATA_MODEL_TYPES_MAPPING.get(value.lower(), str), copy=False)
                if columns:
                    columns = list(set(columns))
                    d = d[list(set(columns))]
                return d

        else:
            raise ValueError(f'backend should be either "dask" or "pandas" you entered {backend}')

        if merge_records:
            if not return_dask_graph:
                d.sort_values('mdmCounterForEntity', inplace=True)
                d.reset_index(inplace=True, drop=True)
                d.drop_duplicates(subset='mdmId', keep='last', inplace=True)
                d.reset_index(inplace=True, drop=True)
            else:
                d = d.set_index('mdmCounterForEntity', sorted=True) \
                    .drop_duplicates(subset='mdmId', keep='last') \
                    .reset_index(drop=True)

        if not return_metadata:
            to_drop = set(['mdmId', 'mdmCounterForEntity', 'mdmLastUpdated']).intersection(set(d.columns))
            d = d.drop(labels=to_drop, axis=1)

        return d


    def export_all(self, sync_view=True, full_export=False, delete_previous=False):
        """

        Export all datamodel to s3

        This method will trigger or pause the export of the data in the data model view to CDS

        :param sync_view: `bool`, default `True`
            Sync the data model view
        :param full_export: `bool`, default `True`
            Do a resync of the data model view
        :param delete_previous: `bool`, default `False`
            Delete previous exported files.
        :return: None
        """
        self.get_all()

        for _name, i in self.template_dict.items():
            view_id = i['mdmId']
            self.export(view_id=view_id, sync_view=sync_view, full_export=full_export,
                        delete_previous=delete_previous)


    def _get_view_export_stats(self):
        """
        Get export status for views

        :return: `dict`
            dict with the information of which data model view is exporting its data.
        """

        json_q = Filter.Builder(key_prefix="") \
            .must(TYPE_FILTER(value="mdmGoldenRecordViewExport")).build().to_json()

        query = Query(self.carol, index_type='CONFIG', page_size=1000, only_hits=False)
        query.query(json_q, ).go()

        dm_results = query.results
        dm_results = [elem.get('hits', elem) for elem in dm_results
                      if elem.get('hits', None)]
        dm_results = list(itertools.chain(*dm_results))

        dm = self.get_all().template_data
        dm = {i['mdmId']: i['mdmName'] for i in dm}

        if dm_results is not None:
            return {dm.get(i['mdmRelationshipViewId'], i['mdmRelationshipViewId'] + '_NOT_FOUND'): i for i in dm_results}

        return dm_results


